@@ -1,25 +1,26 @@
-extern crate tokenizers as tk;
-
-use crate::error::PyError;
+use pyo3::exceptions;
 use pyo3::prelude::*;
 use pyo3::types::*;
 use pyo3::{PyObjectProtocol, PySequenceProtocol};
 use tk::tokenizer::{Offsets, PaddingDirection};
+use tokenizers as tk;
 
-#[pyclass(dict)]
+use crate::error::PyError;
+
+#[pyclass(dict, module = "tokenizers", name=Encoding)]
 #[repr(transparent)]
-pub struct Encoding {
+pub struct PyEncoding {
     pub encoding: tk::tokenizer::Encoding,
 }
 
-impl Encoding {
-    pub fn new(encoding: tk::tokenizer::Encoding) -> Self {
-        Encoding { encoding }
+impl From<tk::tokenizer::Encoding> for PyEncoding {
+    fn from(v: tk::tokenizer::Encoding) -> Self {
+        Self { encoding: v }
     }
 }
 
 #[pyproto]
-impl PyObjectProtocol for Encoding {
+impl PyObjectProtocol for PyEncoding {
     fn __repr__(&self) -> PyResult<String> {
         Ok(format!(
             "Encoding(num_tokens={}, attributes=[ids, type_ids, tokens, offsets, \
@@ -30,25 +31,54 @@ impl PyObjectProtocol for Encoding {
 }
 
 #[pyproto]
-impl PySequenceProtocol for Encoding {
+impl PySequenceProtocol for PyEncoding {
     fn __len__(self) -> PyResult<usize> {
-        Ok(self.encoding.get_ids().len())
+        Ok(self.encoding.len())
     }
 }
 
 #[pymethods]
-impl Encoding {
+impl PyEncoding {
+    #[new]
+    fn new() -> PyResult<Self> {
+        Ok(Self {
+            encoding: tk::tokenizer::Encoding::default(),
+        })
+    }
+
+    fn __getstate__(&self, py: Python) -> PyResult<PyObject> {
+        let data = serde_json::to_string(&self.encoding).map_err(|e| {
+            exceptions::PyException::new_err(format!(
+                "Error while attempting to pickle Encoding: {}",
+                e.to_string()
+            ))
+        })?;
+        Ok(PyBytes::new(py, data.as_bytes()).to_object(py))
+    }
+
+    fn __setstate__(&mut self, py: Python, state: PyObject) -> PyResult<()> {
+        match state.extract::<&PyBytes>(py) {
+            Ok(s) => {
+                self.encoding = serde_json::from_slice(s.as_bytes()).map_err(|e| {
+                    exceptions::PyException::new_err(format!(
+                        "Error while attempting to unpickle Encoding: {}",
+                        e.to_string()
+                    ))
+                })?;
+                Ok(())
+            }
+            Err(e) => Err(e),
+        }
+    }
+
     #[staticmethod]
     #[args(growing_offsets = true)]
-    fn merge(encodings: Vec<PyRef<Encoding>>, growing_offsets: bool) -> Encoding {
-        Encoding::new(tk::tokenizer::Encoding::merge(
-            encodings
-                .into_iter()
-                .map(|e| e.encoding.clone())
-                .collect::<Vec<_>>()
-                .as_slice(),
+    fn merge(encodings: Vec<PyRef<PyEncoding>>, growing_offsets: bool) -> PyEncoding {
+        tk::tokenizer::Encoding::merge(
+            encodings.into_iter().map(|e| e.encoding.clone()),
             growing_offsets,
-        ))
+        )
+        .into()
     }
 
     #[getter]
@@ -87,12 +117,12 @@ impl Encoding {
     }
 
     #[getter]
-    fn get_overflowing(&self) -> Vec<Encoding> {
+    fn get_overflowing(&self) -> Vec<PyEncoding> {
         self.encoding
             .get_overflowing()
             .clone()
             .into_iter()
-            .map(Encoding::new)
+            .map(|e| e.into())
             .collect()
     }
 
@@ -141,7 +171,7 @@ impl Encoding {
                                  one of `left` or `right`",
                                 other
                             ))
-                            .into_pyerr()),
+                            .into_pyerr::<exceptions::PyValueError>()),
                         }?;
                     }
                     "pad_id" => pad_id = value.extract()?,
@@ -151,10 +181,9 @@ impl Encoding {
                 }
             }
         }
-
-        Ok(self
-            .encoding
-            .pad(length, pad_id, pad_type_id, pad_token, direction))
+        self.encoding
+            .pad(length, pad_id, pad_type_id, pad_token, direction);
+        Ok(())
     }
 
     #[args(kwargs = "**")]
@@ -170,7 +199,7 @@ impl Encoding {
                 }
             }
         }
-
-        Ok(self.encoding.truncate(max_length, stride))
+        self.encoding.truncate(max_length, stride);
+        Ok(())
     }
 }
