@@ -1,27 +1,29 @@
-extern crate tokenizers as tk;
+use std::collections::HashMap;
+use std::sync::Arc;
 
+use numpy::PyArray1;
 use pyo3::exceptions;
 use pyo3::prelude::*;
 use pyo3::types::*;
 use pyo3::PyObjectProtocol;
-use std::collections::HashMap;
-
-use super::decoders::Decoder;
-use super::encoding::Encoding;
-use super::error::{PyError, ToPyResult};
-use super::models::Model;
-use super::normalizers::Normalizer;
-use super::pre_tokenizers::PreTokenizer;
-use super::processors::PostProcessor;
-use super::trainers::Trainer;
-use super::utils::Container;
-
+use tk::models::bpe::BPE;
 use tk::tokenizer::{
-    PaddingDirection, PaddingParams, PaddingStrategy, TruncationParams, TruncationStrategy,
+    PaddingDirection, PaddingParams, PaddingStrategy, PostProcessor, TokenizerImpl,
+    TruncationParams, TruncationStrategy,
 };
+use tokenizers as tk;
 
-#[pyclass(dict, module = "tokenizers")]
-pub struct AddedToken {
+use super::decoders::PyDecoder;
+use super::encoding::PyEncoding;
+use super::error::{PyError, ToPyResult};
+use super::models::PyModel;
+use super::normalizers::PyNormalizer;
+use super::pre_tokenizers::PyPreTokenizer;
+use super::trainers::PyTrainer;
+use crate::processors::PyPostProcessor;
+
+#[pyclass(dict, module = "tokenizers", name=AddedToken)]
+pub struct PyAddedToken {
     pub content: String,
     pub is_special_token: bool,
     pub single_word: Option<bool>,
@@ -29,7 +31,7 @@ pub struct AddedToken {
     pub rstrip: Option<bool>,
     pub normalized: Option<bool>,
 }
-impl AddedToken {
+impl PyAddedToken {
     pub fn from<S: Into<String>>(content: S, is_special_token: Option<bool>) -> Self {
         Self {
             content: content.into(),
@@ -75,11 +77,11 @@ impl AddedToken {
 }
 
 #[pymethods]
-impl AddedToken {
+impl PyAddedToken {
     #[new]
     #[args(kwargs = "**")]
     fn new(content: Option<&str>, kwargs: Option<&PyDict>) -> PyResult<Self> {
-        let mut token = AddedToken::from(content.unwrap_or(""), None);
+        let mut token = PyAddedToken::from(content.unwrap_or(""), None);
 
         if let Some(kwargs) = kwargs {
             for (key, value) in kwargs {
@@ -107,6 +109,7 @@ impl AddedToken {
                 for (key, value) in state {
                     let key: &str = key.extract()?;
                     match key {
+                        "content" => self.content = value.extract()?,
                         "single_word" => self.single_word = Some(value.extract()?),
                         "lstrip" => self.lstrip = Some(value.extract()?),
                         "rstrip" => self.rstrip = Some(value.extract()?),
@@ -146,7 +149,7 @@ impl AddedToken {
     }
 }
 #[pyproto]
-impl PyObjectProtocol for AddedToken {
+impl PyObjectProtocol for PyAddedToken {
     fn __str__(&'p self) -> PyResult<&'p str> {
         Ok(&self.content)
     }
@@ -169,126 +172,226 @@ impl PyObjectProtocol for AddedToken {
     }
 }
 
-struct TextInputSequence(tk::InputSequence);
-impl FromPyObject<'_> for TextInputSequence {
-    fn extract(ob: &PyAny) -> PyResult<Self> {
-        let err = exceptions::ValueError::py_err("TextInputSequence must be str");
+struct TextInputSequence<'s>(tk::InputSequence<'s>);
+impl<'s> FromPyObject<'s> for TextInputSequence<'s> {
+    fn extract(ob: &'s PyAny) -> PyResult<Self> {
+        let err = exceptions::PyTypeError::new_err("TextInputSequence must be str");
         if let Ok(s) = ob.downcast::<PyString>() {
-            let seq: String = s.extract().map_err(|_| err)?;
-            Ok(Self(seq.into()))
+            Ok(Self(s.to_string_lossy().into()))
         } else {
             Err(err)
         }
     }
 }
-impl From<TextInputSequence> for tk::InputSequence {
-    fn from(s: TextInputSequence) -> Self {
+impl<'s> From<TextInputSequence<'s>> for tk::InputSequence<'s> {
+    fn from(s: TextInputSequence<'s>) -> Self {
         s.0
     }
 }
 
-struct PreTokenizedInputSequence(tk::InputSequence);
-impl FromPyObject<'_> for PreTokenizedInputSequence {
+struct PyArrayUnicode(Vec<String>);
+impl FromPyObject<'_> for PyArrayUnicode {
     fn extract(ob: &PyAny) -> PyResult<Self> {
-        let err = exceptions::ValueError::py_err(
-            "PreTokenizedInputSequence must be Union[List[str], Tuple[str]]",
-        );
+        let array = ob.downcast::<PyArray1<u8>>()?;
+        let arr = array.as_array_ptr();
+        let (type_num, elsize, alignment, data) = unsafe {
+            let desc = (*arr).descr;
+            (
+                (*desc).type_num,
+                (*desc).elsize as usize,
+                (*desc).alignment as usize,
+                (*arr).data,
+            )
+        };
+        let n_elem = array.shape()[0];
 
+        // type_num == 19 => Unicode
+        if type_num != 19 {
+            return Err(exceptions::PyTypeError::new_err(
+                "Expected a np.array[dtype='U']",
+            ));
+        }
+
+        unsafe {
+            let all_bytes = std::slice::from_raw_parts(data as *const u8, elsize * n_elem);
+
+            let seq = (0..n_elem)
+                .map(|i| {
+                    let bytes = &all_bytes[i * elsize..(i + 1) * elsize];
+                    let unicode = pyo3::ffi::PyUnicode_FromUnicode(
+                        bytes.as_ptr() as *const _,
+                        elsize as isize / alignment as isize,
+                    );
+                    let gil = Python::acquire_gil();
+                    let py = gil.python();
+                    let obj = PyObject::from_owned_ptr(py, unicode);
+                    let s = obj.cast_as::<PyString>(py)?;
+                    Ok(s.to_string_lossy().trim_matches(char::from(0)).to_owned())
+                })
+                .collect::<PyResult<Vec<_>>>()?;
+
+            Ok(Self(seq))
+        }
+    }
+}
+impl From<PyArrayUnicode> for tk::InputSequence<'_> {
+    fn from(s: PyArrayUnicode) -> Self {
+        s.0.into()
+    }
+}
+
+struct PyArrayStr(Vec<String>);
+impl FromPyObject<'_> for PyArrayStr {
+    fn extract(ob: &PyAny) -> PyResult<Self> {
+        let array = ob.downcast::<PyArray1<u8>>()?;
+        let arr = array.as_array_ptr();
+        let (type_num, data) = unsafe { ((*(*arr).descr).type_num, (*arr).data) };
+        let n_elem = array.shape()[0];
+
+        if type_num != 17 {
+            return Err(exceptions::PyTypeError::new_err(
+                "Expected a np.array[dtype='O']",
+            ));
+        }
+
+        unsafe {
+            let objects = std::slice::from_raw_parts(data as *const PyObject, n_elem);
+
+            let seq = objects
+                .iter()
+                .map(|obj| {
+                    let gil = Python::acquire_gil();
+                    let py = gil.python();
+                    let s = obj.cast_as::<PyString>(py)?;
+                    Ok(s.to_string_lossy().into_owned())
+                })
+                .collect::<PyResult<Vec<_>>>()?;
+
+            Ok(Self(seq))
+        }
+    }
+}
+impl From<PyArrayStr> for tk::InputSequence<'_> {
+    fn from(s: PyArrayStr) -> Self {
+        s.0.into()
+    }
+}
+
+struct PreTokenizedInputSequence<'s>(tk::InputSequence<'s>);
+impl<'s> FromPyObject<'s> for PreTokenizedInputSequence<'s> {
+    fn extract(ob: &'s PyAny) -> PyResult<Self> {
+        if let Ok(seq) = ob.extract::<PyArrayUnicode>() {
+            return Ok(Self(seq.into()));
+        }
+        if let Ok(seq) = ob.extract::<PyArrayStr>() {
+            return Ok(Self(seq.into()));
+        }
         if let Ok(s) = ob.downcast::<PyList>() {
-            let seq = s.extract::<Vec<String>>().map_err(|_| err)?;
-            Ok(Self(seq.into()))
-        } else if let Ok(s) = ob.downcast::<PyTuple>() {
-            let seq = s.extract::<Vec<String>>().map_err(|_| err)?;
-            Ok(Self(seq.into()))
-        } else {
-            Err(err)
+            if let Ok(seq) = s.extract::<Vec<&str>>() {
+                return Ok(Self(seq.into()));
+            }
         }
+        if let Ok(s) = ob.downcast::<PyTuple>() {
+            if let Ok(seq) = s.extract::<Vec<&str>>() {
+                return Ok(Self(seq.into()));
+            }
+        }
+        Err(exceptions::PyTypeError::new_err(
+            "PreTokenizedInputSequence must be Union[List[str], Tuple[str]]",
+        ))
     }
 }
-impl From<PreTokenizedInputSequence> for tk::InputSequence {
-    fn from(s: PreTokenizedInputSequence) -> Self {
+impl<'s> From<PreTokenizedInputSequence<'s>> for tk::InputSequence<'s> {
+    fn from(s: PreTokenizedInputSequence<'s>) -> Self {
         s.0
     }
 }
 
-struct TextEncodeInput(tk::EncodeInput);
-impl FromPyObject<'_> for TextEncodeInput {
-    fn extract(ob: &PyAny) -> PyResult<Self> {
-        let err = exceptions::ValueError::py_err(
-            "TextEncodeInput must be Union[TextInputSequence, Tuple[InputSequence, InputSequence]]",
-        );
-
-        let gil = Python::acquire_gil();
-        let py = gil.python();
-        let obj = ob.to_object(py);
-
-        if let Ok(i) = obj.extract::<TextInputSequence>(py) {
-            Ok(Self(i.into()))
-        } else if let Ok((i1, i2)) = obj.extract::<(TextInputSequence, TextInputSequence)>(py) {
-            Ok(Self((i1, i2).into()))
-        } else {
-            Err(err)
+struct TextEncodeInput<'s>(tk::EncodeInput<'s>);
+impl<'s> FromPyObject<'s> for TextEncodeInput<'s> {
+    fn extract(ob: &'s PyAny) -> PyResult<Self> {
+        if let Ok(i) = ob.extract::<TextInputSequence>() {
+            return Ok(Self(i.into()));
         }
+        if let Ok((i1, i2)) = ob.extract::<(TextInputSequence, TextInputSequence)>() {
+            return Ok(Self((i1, i2).into()));
+        }
+        if let Ok(arr) = ob.extract::<Vec<&PyAny>>() {
+            if arr.len() == 2 {
+                let first = arr[0].extract::<TextInputSequence>()?;
+                let second = arr[1].extract::<TextInputSequence>()?;
+                return Ok(Self((first, second).into()));
+            }
+        }
+        Err(exceptions::PyTypeError::new_err(
+            "TextEncodeInput must be Union[TextInputSequence, Tuple[InputSequence, InputSequence]]",
+        ))
     }
 }
-impl From<TextEncodeInput> for tk::tokenizer::EncodeInput {
-    fn from(i: TextEncodeInput) -> Self {
+impl<'s> From<TextEncodeInput<'s>> for tk::tokenizer::EncodeInput<'s> {
+    fn from(i: TextEncodeInput<'s>) -> Self {
         i.0
     }
 }
-struct PreTokenizedEncodeInput(tk::EncodeInput);
-impl FromPyObject<'_> for PreTokenizedEncodeInput {
-    fn extract(ob: &PyAny) -> PyResult<Self> {
-        let err = exceptions::ValueError::py_err(
+struct PreTokenizedEncodeInput<'s>(tk::EncodeInput<'s>);
+impl<'s> FromPyObject<'s> for PreTokenizedEncodeInput<'s> {
+    fn extract(ob: &'s PyAny) -> PyResult<Self> {
+        if let Ok(i) = ob.extract::<PreTokenizedInputSequence>() {
+            return Ok(Self(i.into()));
+        }
+        if let Ok((i1, i2)) = ob.extract::<(PreTokenizedInputSequence, PreTokenizedInputSequence)>()
+        {
+            return Ok(Self((i1, i2).into()));
+        }
+        if let Ok(arr) = ob.extract::<Vec<&PyAny>>() {
+            if arr.len() == 2 {
+                let first = arr[0].extract::<PreTokenizedInputSequence>()?;
+                let second = arr[1].extract::<PreTokenizedInputSequence>()?;
+                return Ok(Self((first, second).into()));
+            }
+        }
+        Err(exceptions::PyTypeError::new_err(
             "PreTokenizedEncodeInput must be Union[PreTokenizedInputSequence, \
             Tuple[PreTokenizedInputSequence, PreTokenizedInputSequence]]",
-        );
-
-        let gil = Python::acquire_gil();
-        let py = gil.python();
-        let obj = ob.to_object(py);
-
-        if let Ok(i) = obj.extract::<PreTokenizedInputSequence>(py) {
-            Ok(Self(i.into()))
-        } else if let Ok((i1, i2)) =
-            obj.extract::<(PreTokenizedInputSequence, PreTokenizedInputSequence)>(py)
-        {
-            Ok(Self((i1, i2).into()))
-        } else {
-            Err(err)
-        }
+        ))
     }
 }
-impl From<PreTokenizedEncodeInput> for tk::tokenizer::EncodeInput {
-    fn from(i: PreTokenizedEncodeInput) -> Self {
+impl<'s> From<PreTokenizedEncodeInput<'s>> for tk::tokenizer::EncodeInput<'s> {
+    fn from(i: PreTokenizedEncodeInput<'s>) -> Self {
         i.0
     }
 }
 
-#[pyclass(dict, module = "tokenizers")]
-pub struct Tokenizer {
-    tokenizer: tk::tokenizer::Tokenizer,
+type Tokenizer = TokenizerImpl<PyModel, PyNormalizer, PyPreTokenizer, PyPostProcessor, PyDecoder>;
+
+#[pyclass(dict, module = "tokenizers", name=Tokenizer)]
+#[derive(Clone)]
+pub struct PyTokenizer {
+    tokenizer: Tokenizer,
+}
+
+impl PyTokenizer {
+    fn new(tokenizer: Tokenizer) -> Self {
+        PyTokenizer { tokenizer }
+    }
+
+    fn from_model(model: PyModel) -> Self {
+        PyTokenizer::new(TokenizerImpl::new(model))
+    }
 }
 
 #[pymethods]
-impl Tokenizer {
+impl PyTokenizer {
     #[new]
-    fn new(mut model: PyRefMut<Model>) -> PyResult<Self> {
-        if let Some(model) = model.model.to_pointer() {
-            let tokenizer = tk::tokenizer::Tokenizer::new(model);
-            Ok(Tokenizer { tokenizer })
-        } else {
-            Err(exceptions::Exception::py_err(
-                "The Model is already being used in another Tokenizer",
-            ))
-        }
+    fn __new__(model: PyRef<PyModel>) -> Self {
+        PyTokenizer::from_model(model.clone())
     }
 
     fn __getstate__(&self, py: Python) -> PyResult<PyObject> {
         let data = serde_json::to_string(&self.tokenizer).map_err(|e| {
-            exceptions::Exception::py_err(format!(
+            exceptions::PyException::new_err(format!(
                 "Error while attempting to pickle Tokenizer: {}",
-                e.to_string()
+                e
             ))
         })?;
         Ok(PyBytes::new(py, data.as_bytes()).to_object(py))
@@ -298,9 +401,9 @@ impl Tokenizer {
         match state.extract::<&PyBytes>(py) {
             Ok(s) => {
                 self.tokenizer = serde_json::from_slice(s.as_bytes()).map_err(|e| {
-                    exceptions::Exception::py_err(format!(
+                    exceptions::PyException::new_err(format!(
                         "Error while attempting to unpickle Tokenizer: {}",
-                        e.to_string()
+                        e
                     ))
                 })?;
                 Ok(())
@@ -310,39 +413,31 @@ impl Tokenizer {
     }
 
     fn __getnewargs__<'p>(&self, py: Python<'p>) -> PyResult<&'p PyTuple> {
-        let model: PyObject = crate::models::Model {
-            model: Container::Owned(Box::new(tk::models::bpe::BPE::default())),
-        }
-        .into_py(py);
+        let model: PyObject = PyModel::new(Arc::new(BPE::default().into())).into_py(py);
         let args = PyTuple::new(py, vec![model]);
         Ok(args)
     }
 
     #[staticmethod]
     fn from_str(s: &str) -> PyResult<Self> {
-        let tokenizer: PyResult<tk::tokenizer::Tokenizer> = ToPyResult(s.parse()).into();
-        Ok(Self {
-            tokenizer: tokenizer?,
-        })
+        let tokenizer: PyResult<_> = ToPyResult(s.parse()).into();
+        Ok(Self::new(tokenizer?))
     }
 
     #[staticmethod]
     fn from_file(path: &str) -> PyResult<Self> {
-        let tokenizer: PyResult<_> = ToPyResult(tk::tokenizer::Tokenizer::from_file(path)).into();
-        Ok(Self {
-            tokenizer: tokenizer?,
-        })
+        let tokenizer: PyResult<_> = ToPyResult(Tokenizer::from_file(path)).into();
+        Ok(Self::new(tokenizer?))
     }
 
     #[staticmethod]
     fn from_buffer(buffer: &PyBytes) -> PyResult<Self> {
-        let tokenizer: tk::tokenizer::Tokenizer = serde_json::from_slice(buffer.as_bytes())
-            .map_err(|e| {
-                exceptions::Exception::py_err(format!(
-                    "Cannot instantiate Tokenizer from buffer: {}",
-                    e.to_string()
-                ))
-            })?;
+        let tokenizer = serde_json::from_slice(buffer.as_bytes()).map_err(|e| {
+            exceptions::PyValueError::new_err(format!(
+                "Cannot instantiate Tokenizer from buffer: {}",
+                e
+            ))
+        })?;
         Ok(Self { tokenizer })
     }
 
@@ -360,7 +455,7 @@ impl Tokenizer {
         Ok(self
             .tokenizer
             .get_post_processor()
-            .map_or(0, |p| p.as_ref().added_tokens(is_pair)))
+            .map_or(0, |p| p.added_tokens(is_pair)))
     }
 
     #[args(with_added_tokens = true)]
@@ -394,7 +489,7 @@ impl Tokenizer {
                                  one of `longest_first`, `only_first`, or `only_second`",
                                 value
                             ))
-                            .into_pyerr()),
+                            .into_pyerr::<exceptions::PyValueError>()),
                         }?
                     }
                     _ => println!("Ignored unknown kwarg option {}", key),
@@ -442,7 +537,7 @@ impl Tokenizer {
                                  one of `left` or `right`",
                                 other
                             ))
-                            .into_pyerr()),
+                            .into_pyerr::<exceptions::PyValueError>()),
                         }?;
                     }
                     "pad_to_multiple_of" => {
@@ -507,15 +602,6 @@ impl Tokenizer {
         })
     }
 
-    fn normalize(&self, sentence: &str) -> PyResult<String> {
-        ToPyResult(
-            self.tokenizer
-                .normalize(sentence)
-                .map(|s| s.get().to_owned()),
-        )
-        .into()
-    }
-
     /// Input can be:
     /// encode("A single sequence")
     /// encode("A sequence", "And its pair")
@@ -531,7 +617,7 @@ impl Tokenizer {
         pair: Option<&PyAny>,
         is_pretokenized: bool,
         add_special_tokens: bool,
-    ) -> PyResult<Encoding> {
+    ) -> PyResult<PyEncoding> {
         let sequence: tk::InputSequence = if is_pretokenized {
             sequence.extract::<PreTokenizedInputSequence>()?.into()
         } else {
@@ -551,7 +637,7 @@ impl Tokenizer {
 
         ToPyResult(
             self.tokenizer
-                .encode(input, add_special_tokens)
+                .encode_char_offsets(input, add_special_tokens)
                 .map(|e| e.into()),
         )
         .into()
@@ -570,7 +656,7 @@ impl Tokenizer {
         input: Vec<&PyAny>,
         is_pretokenized: bool,
         add_special_tokens: bool,
-    ) -> PyResult<Vec<Encoding>> {
+    ) -> PyResult<Vec<PyEncoding>> {
         let input: Vec<tk::EncodeInput> = input
             .into_iter()
             .map(|o| {
@@ -582,12 +668,15 @@ impl Tokenizer {
                 Ok(input)
             })
             .collect::<PyResult<Vec<tk::EncodeInput>>>()?;
-        ToPyResult(
-            self.tokenizer
-                .encode_batch(input, add_special_tokens)
-                .map(|encodings| encodings.into_iter().map(|e| e.into()).collect()),
-        )
-        .into()
+        let gil = Python::acquire_gil();
+        gil.python().allow_threads(|| {
+            ToPyResult(
+                self.tokenizer
+                    .encode_batch_char_offsets(input, add_special_tokens)
+                    .map(|encodings| encodings.into_iter().map(|e| e.into()).collect()),
+            )
+            .into()
+        })
     }
 
     fn decode(&self, ids: Vec<u32>, skip_special_tokens: Option<bool>) -> PyResult<String> {
@@ -603,11 +692,14 @@ impl Tokenizer {
         sentences: Vec<Vec<u32>>,
         skip_special_tokens: Option<bool>,
     ) -> PyResult<Vec<String>> {
-        ToPyResult(
-            self.tokenizer
-                .decode_batch(sentences, skip_special_tokens.unwrap_or(true)),
-        )
-        .into()
+        let gil = Python::acquire_gil();
+        gil.python().allow_threads(|| {
+            ToPyResult(
+                self.tokenizer
+                    .decode_batch(sentences, skip_special_tokens.unwrap_or(true)),
+            )
+            .into()
+        })
     }
 
     fn token_to_id(&self, token: &str) -> Option<u32> {
@@ -623,12 +715,12 @@ impl Tokenizer {
             .into_iter()
             .map(|token| {
                 if let Ok(content) = token.extract::<String>() {
-                    Ok(AddedToken::from(content, Some(false)).get_token())
-                } else if let Ok(mut token) = token.extract::<PyRefMut<AddedToken>>() {
+                    Ok(PyAddedToken::from(content, Some(false)).get_token())
+                } else if let Ok(mut token) = token.extract::<PyRefMut<PyAddedToken>>() {
                     token.is_special_token = false;
                     Ok(token.get_token())
                 } else {
-                    Err(exceptions::Exception::py_err(
+                    Err(exceptions::PyTypeError::new_err(
                         "Input must be a List[Union[str, AddedToken]]",
                     ))
                 }
@@ -644,11 +736,11 @@ impl Tokenizer {
             .map(|token| {
                 if let Ok(content) = token.extract::<String>() {
                     Ok(tk::tokenizer::AddedToken::from(content, true))
-                } else if let Ok(mut token) = token.extract::<PyRefMut<AddedToken>>() {
+                } else if let Ok(mut token) = token.extract::<PyRefMut<PyAddedToken>>() {
                     token.is_special_token = true;
                     Ok(token.get_token())
                 } else {
-                    Err(exceptions::Exception::py_err(
+                    Err(exceptions::PyTypeError::new_err(
                         "Input must be a List[Union[str, AddedToken]]",
                     ))
                 }
@@ -658,23 +750,17 @@ impl Tokenizer {
         Ok(self.tokenizer.add_special_tokens(&tokens))
     }
 
-    fn train(&mut self, trainer: &Trainer, files: Vec<String>) -> PyResult<()> {
-        trainer.trainer.execute(|trainer| {
-            if let Err(e) = self.tokenizer.train(trainer, files) {
-                Err(exceptions::Exception::py_err(format!("{}", e)))
-            } else {
-                Ok(())
-            }
-        })
+    fn train(&mut self, trainer: &PyTrainer, files: Vec<String>) -> PyResult<()> {
+        ToPyResult(self.tokenizer.train_and_replace(trainer, files)).into()
     }
 
     #[args(pair = "None", add_special_tokens = true)]
     fn post_process(
         &self,
-        encoding: &Encoding,
-        pair: Option<&Encoding>,
+        encoding: &PyEncoding,
+        pair: Option<&PyEncoding>,
         add_special_tokens: bool,
-    ) -> PyResult<Encoding> {
+    ) -> PyResult<PyEncoding> {
         ToPyResult(
             self.tokenizer
                 .post_process(
@@ -688,106 +774,94 @@ impl Tokenizer {
     }
 
     #[getter]
-    fn get_model(&self) -> PyResult<Model> {
-        Ok(Model {
-            model: Container::from_ref(self.tokenizer.get_model()),
-        })
+    fn get_model(&self) -> PyResult<PyObject> {
+        self.tokenizer.get_model().get_as_subtype()
     }
 
     #[setter]
-    fn set_model(&mut self, mut model: PyRefMut<Model>) -> PyResult<()> {
-        if let Some(model) = model.model.to_pointer() {
-            self.tokenizer.with_model(model);
-            Ok(())
-        } else {
-            Err(exceptions::Exception::py_err(
-                "The Model is already being used in another Tokenizer",
-            ))
-        }
+    fn set_model(&mut self, model: PyRef<PyModel>) {
+        self.tokenizer.with_model(model.clone());
     }
 
     #[getter]
-    fn get_normalizer(&self) -> PyResult<Option<Normalizer>> {
-        Ok(self
-            .tokenizer
-            .get_normalizer()
-            .map(|normalizer| Normalizer {
-                normalizer: Container::from_ref(normalizer),
-            }))
+    fn get_normalizer(&self) -> PyResult<PyObject> {
+        if let Some(n) = self.tokenizer.get_normalizer() {
+            n.get_as_subtype()
+        } else {
+            Ok(Python::acquire_gil().python().None())
+        }
     }
 
     #[setter]
-    fn set_normalizer(&mut self, mut normalizer: PyRefMut<Normalizer>) -> PyResult<()> {
-        if let Some(normalizer) = normalizer.normalizer.to_pointer() {
-            self.tokenizer.with_normalizer(normalizer);
-            Ok(())
-        } else {
-            Err(exceptions::Exception::py_err(
-                "The Normalizer is already being used in another Tokenizer",
-            ))
-        }
+    fn set_normalizer(&mut self, normalizer: PyRef<PyNormalizer>) {
+        self.tokenizer.with_normalizer(normalizer.clone());
     }
 
     #[getter]
-    fn get_pre_tokenizer(&self) -> PyResult<Option<PreTokenizer>> {
-        Ok(self
-            .tokenizer
-            .get_pre_tokenizer()
-            .map(|pretok| PreTokenizer {
-                pretok: Container::from_ref(pretok),
-            }))
+    fn get_pre_tokenizer(&self) -> PyResult<PyObject> {
+        if let Some(pt) = self.tokenizer.get_pre_tokenizer() {
+            pt.get_as_subtype()
+        } else {
+            Ok(Python::acquire_gil().python().None())
+        }
     }
 
     #[setter]
-    fn set_pre_tokenizer(&mut self, mut pretok: PyRefMut<PreTokenizer>) -> PyResult<()> {
-        if let Some(pretok) = pretok.pretok.to_pointer() {
-            self.tokenizer.with_pre_tokenizer(pretok);
-            Ok(())
-        } else {
-            Err(exceptions::Exception::py_err(
-                "The PreTokenizer is already being used in another Tokenizer",
-            ))
-        }
+    fn set_pre_tokenizer(&mut self, pretok: PyRef<PyPreTokenizer>) {
+        self.tokenizer.with_pre_tokenizer(pretok.clone());
     }
 
     #[getter]
-    fn get_post_processor(&self) -> PyResult<Option<PostProcessor>> {
-        Ok(self
-            .tokenizer
-            .get_post_processor()
-            .map(|processor| PostProcessor {
-                processor: Container::from_ref(processor),
-            }))
+    fn get_post_processor(&self) -> PyResult<PyObject> {
+        if let Some(n) = self.tokenizer.get_post_processor() {
+            n.get_as_subtype()
+        } else {
+            Ok(Python::acquire_gil().python().None())
+        }
     }
 
     #[setter]
-    fn set_post_processor(&mut self, mut processor: PyRefMut<PostProcessor>) -> PyResult<()> {
-        if let Some(processor) = processor.processor.to_pointer() {
-            self.tokenizer.with_post_processor(processor);
-            Ok(())
-        } else {
-            Err(exceptions::Exception::py_err(
-                "The Processor is already being used in another Tokenizer",
-            ))
-        }
+    fn set_post_processor(&mut self, processor: PyRef<PyPostProcessor>) {
+        self.tokenizer.with_post_processor(processor.clone());
     }
 
     #[getter]
-    fn get_decoder(&self) -> PyResult<Option<Decoder>> {
-        Ok(self.tokenizer.get_decoder().map(|decoder| Decoder {
-            decoder: Container::from_ref(decoder),
-        }))
+    fn get_decoder(&self) -> PyResult<PyObject> {
+        if let Some(dec) = self.tokenizer.get_decoder() {
+            dec.get_as_subtype()
+        } else {
+            Ok(Python::acquire_gil().python().None())
+        }
     }
 
     #[setter]
-    fn set_decoder(&mut self, mut decoder: PyRefMut<Decoder>) -> PyResult<()> {
-        if let Some(decoder) = decoder.decoder.to_pointer() {
-            self.tokenizer.with_decoder(decoder);
-            Ok(())
-        } else {
-            Err(exceptions::Exception::py_err(
-                "The Decoder is already being used in another Tokenizer",
-            ))
-        }
+    fn set_decoder(&mut self, decoder: PyRef<PyDecoder>) {
+        self.tokenizer.with_decoder(decoder.clone());
+    }
+}
+
+#[cfg(test)]
+mod test {
+    use super::*;
+    use crate::models::PyModel;
+    use crate::normalizers::{PyNormalizer, PyNormalizerTypeWrapper};
+    use std::sync::Arc;
+    use tempfile::NamedTempFile;
+    use tk::normalizers::{Lowercase, NFKC};
+
+    #[test]
+    fn serialize() {
+        let mut tokenizer = Tokenizer::new(PyModel::new(Arc::new(
+            tk::models::bpe::BPE::default().into(),
+        )));
+        tokenizer.with_normalizer(PyNormalizer::new(PyNormalizerTypeWrapper::Sequence(vec![
+            Arc::new(NFKC.into()),
+            Arc::new(Lowercase.into()),
+        ])));
+
+        let tmp = NamedTempFile::new().unwrap().into_temp_path();
+        tokenizer.save(&tmp, false).unwrap();
+
+        Tokenizer::from_file(&tmp).unwrap();
     }
 }
