@@ -1,27 +1,57 @@
-extern crate tokenizers as tk;
+use std::collections::HashMap;
+use std::sync::Arc;
 
-use super::utils::Container;
-use crate::tokenizer::AddedToken;
 use pyo3::exceptions;
 use pyo3::prelude::*;
 use pyo3::types::*;
+use tk::models::TrainerWrapper;
+use tk::Trainer;
+use tokenizers as tk;
 
-#[pyclass]
-pub struct Trainer {
-    pub trainer: Container<dyn tk::tokenizer::Trainer>,
+use crate::models::PyModel;
+use crate::tokenizer::PyAddedToken;
+
+#[pyclass(name=Trainer)]
+pub struct PyTrainer {
+    pub trainer: TrainerWrapper,
 }
 
-#[pyclass(extends=Trainer)]
-pub struct BpeTrainer {}
+impl PyTrainer {
+    pub fn new(trainer: TrainerWrapper) -> Self {
+        PyTrainer { trainer }
+    }
+}
+
+impl Trainer for PyTrainer {
+    type Model = PyModel;
+
+    fn should_show_progress(&self) -> bool {
+        self.trainer.should_show_progress()
+    }
+
+    fn train(&self, words: HashMap<String, u32>) -> tk::Result<(PyModel, Vec<tk::AddedToken>)> {
+        self.trainer.train(words).map(|(m, t)| {
+            let m = PyModel { model: Arc::new(m) };
+            (m, t)
+        })
+    }
+
+    fn process_tokens(&self, words: &mut HashMap<String, u32>, tokens: Vec<String>) {
+        self.trainer.process_tokens(words, tokens)
+    }
+}
+
+#[pyclass(extends=PyTrainer, name=BpeTrainer)]
+pub struct PyBpeTrainer {}
 #[pymethods]
-impl BpeTrainer {
+impl PyBpeTrainer {
     /// new(/ vocab_size, min_frequency)
     /// --
     ///
     /// Create a new BpeTrainer with the given configuration
     #[new]
     #[args(kwargs = "**")]
-    pub fn new(kwargs: Option<&PyDict>) -> PyResult<(Self, Trainer)> {
+    pub fn new(kwargs: Option<&PyDict>) -> PyResult<(Self, PyTrainer)> {
         let mut builder = tk::models::bpe::BpeTrainer::builder();
         if let Some(kwargs) = kwargs {
             for (key, val) in kwargs {
@@ -36,14 +66,14 @@ impl BpeTrainer {
                                 .into_iter()
                                 .map(|token| {
                                     if let Ok(content) = token.extract::<String>() {
-                                        Ok(tk::tokenizer::AddedToken {
-                                            content,
-                                            ..Default::default()
-                                        })
-                                    } else if let Ok(token) = token.extract::<PyRef<AddedToken>>() {
-                                        Ok(token.token.clone())
+                                        Ok(PyAddedToken::from(content, Some(true)).get_token())
+                                    } else if let Ok(mut token) =
+                                        token.extract::<PyRefMut<PyAddedToken>>()
+                                    {
+                                        token.is_special_token = true;
+                                        Ok(token.get_token())
                                     } else {
-                                        Err(exceptions::Exception::py_err(
+                                        Err(exceptions::PyTypeError::new_err(
                                             "special_tokens must be a List[Union[str, AddedToken]]",
                                         ))
                                     }
@@ -57,7 +87,7 @@ impl BpeTrainer {
                         builder = builder.initial_alphabet(
                             alphabet
                                 .into_iter()
-                                .map(|s| s.chars().nth(0))
+                                .map(|s| s.chars().next())
                                 .filter(|c| c.is_some())
                                 .map(|c| c.unwrap())
                                 .collect(),
@@ -71,26 +101,21 @@ impl BpeTrainer {
                 };
             }
         }
-        Ok((
-            BpeTrainer {},
-            Trainer {
-                trainer: Container::Owned(Box::new(builder.build())),
-            },
-        ))
+        Ok((PyBpeTrainer {}, PyTrainer::new(builder.build().into())))
     }
 }
 
-#[pyclass(extends=Trainer)]
-pub struct WordPieceTrainer {}
+#[pyclass(extends=PyTrainer, name=WordPieceTrainer)]
+pub struct PyWordPieceTrainer {}
 #[pymethods]
-impl WordPieceTrainer {
+impl PyWordPieceTrainer {
     /// new(/ vocab_size, min_frequency)
     /// --
     ///
     /// Create a new BpeTrainer with the given configuration
     #[new]
     #[args(kwargs = "**")]
-    pub fn new(kwargs: Option<&PyDict>) -> PyResult<(Self, Trainer)> {
+    pub fn new(kwargs: Option<&PyDict>) -> PyResult<(Self, PyTrainer)> {
         let mut builder = tk::models::wordpiece::WordPieceTrainer::builder();
         if let Some(kwargs) = kwargs {
             for (key, val) in kwargs {
@@ -105,14 +130,14 @@ impl WordPieceTrainer {
                                 .into_iter()
                                 .map(|token| {
                                     if let Ok(content) = token.extract::<String>() {
-                                        Ok(tk::tokenizer::AddedToken {
-                                            content,
-                                            ..Default::default()
-                                        })
-                                    } else if let Ok(token) = token.extract::<PyRef<AddedToken>>() {
-                                        Ok(token.token.clone())
+                                        Ok(PyAddedToken::from(content, Some(true)).get_token())
+                                    } else if let Ok(mut token) =
+                                        token.extract::<PyRefMut<PyAddedToken>>()
+                                    {
+                                        token.is_special_token = true;
+                                        Ok(token.get_token())
                                     } else {
-                                        Err(exceptions::Exception::py_err(
+                                        Err(exceptions::PyTypeError::new_err(
                                             "special_tokens must be a List[Union[str, AddedToken]]",
                                         ))
                                     }
@@ -126,7 +151,7 @@ impl WordPieceTrainer {
                         builder = builder.initial_alphabet(
                             alphabet
                                 .into_iter()
-                                .map(|s| s.chars().nth(0))
+                                .map(|s| s.chars().next())
                                 .filter(|c| c.is_some())
                                 .map(|c| c.unwrap())
                                 .collect(),
@@ -142,10 +167,63 @@ impl WordPieceTrainer {
         }
 
         Ok((
-            WordPieceTrainer {},
-            Trainer {
-                trainer: Container::Owned(Box::new(builder.build())),
-            },
+            PyWordPieceTrainer {},
+            PyTrainer::new(builder.build().into()),
         ))
     }
 }
+
+#[pyclass(extends=PyTrainer, name=UnigramTrainer)]
+pub struct PyUnigramTrainer {}
+#[pymethods]
+impl PyUnigramTrainer {
+    /// Create a new UnigramTrainer with the given configuration
+    #[new]
+    #[args(kwargs = "**")]
+    pub fn new(kwargs: Option<&PyDict>) -> PyResult<(Self, PyTrainer)> {
+        let mut builder = tk::models::unigram::UnigramTrainer::builder();
+        if let Some(kwargs) = kwargs {
+            for (key, val) in kwargs {
+                let key: &str = key.extract()?;
+                match key {
+                    "vocab_size" => builder.vocab_size(val.extract()?),
+                    "show_progress" => builder.show_progress(val.extract()?),
+                    "n_sub_iterations" => builder.n_sub_iterations(val.extract()?),
+                    "shrinking_factor" => builder.shrinking_factor(val.extract()?),
+                    "unk_token" => builder.unk_token(val.extract()?),
+                    "max_piece_length" => builder.max_piece_length(val.extract()?),
+                    "seed_size" => builder.seed_size(val.extract()?),
+                    "special_tokens" => builder.special_tokens(
+                        val.cast_as::<PyList>()?
+                            .into_iter()
+                            .map(|token| {
+                                if let Ok(content) = token.extract::<String>() {
+                                    Ok(PyAddedToken::from(content, Some(true)).get_token())
+                                } else if let Ok(mut token) =
+                                    token.extract::<PyRefMut<PyAddedToken>>()
+                                {
+                                    token.is_special_token = true;
+                                    Ok(token.get_token())
+                                } else {
+                                    Err(exceptions::PyTypeError::new_err(
+                                        "special_tokens must be a List[Union[str, AddedToken]]",
+                                    ))
+                                }
+                            })
+                            .collect::<PyResult<Vec<_>>>()?,
+                    ),
+                    _ => {
+                        println!("Ignored unknown kwargs option {}", key);
+                        &mut builder
+                    }
+                };
+            }
+        }
+
+        let trainer: tokenizers::models::unigram::UnigramTrainer =
+            builder.build().map_err(|e| {
+                exceptions::PyException::new_err(format!("Cannot build UnigramTrainer: {}", e))
+            })?;
+        Ok((PyUnigramTrainer {}, PyTrainer::new(trainer.into())))
+    }
+}
