@@ -1,10 +1,12 @@
-use crate::parallelism::*;
-use crate::tokenizer::{
-    Decoder, Encoding, NormalizedString, Offsets, PostProcessor, PreTokenizer, Result,
-};
+use std::collections::{HashMap, HashSet};
+
 use onig::Regex;
 use serde::{Deserialize, Serialize};
-use std::collections::{HashMap, HashSet};
+
+use crate::tokenizer::{
+    Decoder, Encoding, PostProcessor, PreTokenizedString, PreTokenizer, Result,
+    SplitDelimiterBehavior,
+};
 
 fn bytes_char() -> HashMap<u8, char> {
     let mut bs: Vec<u8> = vec![];
@@ -38,10 +40,11 @@ lazy_static! {
         bytes_char().into_iter().map(|(c, b)| (b, c)).collect();
 }
 
-#[derive(Serialize, Deserialize)]
+#[derive(Deserialize, Serialize, Copy, Clone, Debug)]
 /// Provides all the necessary steps to handle the BPE tokenization at the byte-level. Takes care
 /// of all the required processing steps to transform a UTF-8 string as needed before and after the
 /// BPE model does its job.
+#[serde(tag = "type")]
 pub struct ByteLevel {
     /// Whether to add a leading space to the first word. This allows to treat the leading word
     /// just as any other word.
@@ -84,74 +87,38 @@ impl ByteLevel {
 /// As a `PreTokenizer`, `ByteLevel` is in charge of transforming all the unicode characters into
 /// their byte-level counterpart. It also splits the input according to the configured regex.
 // TODO: Give the ability to modify this regex
-#[typetag::serde]
 impl PreTokenizer for ByteLevel {
-    fn pre_tokenize(&self, normalized: &mut NormalizedString) -> Result<Vec<(String, Offsets)>> {
-        if self.add_prefix_space && !normalized.get().starts_with(' ') {
-            normalized.prepend(" ");
-        }
-
-        let positions = RE
-            .find_iter(normalized.get())
-            .map(|(start, end)| start..end)
-            .collect::<Vec<_>>();
-
-        let splits = positions
-            .into_maybe_par_iter()
-            .map(|range| {
-                // Process one of the splits
-                let slice = &normalized.get()[range];
-                let mut chars: Vec<(char, u8)> = Vec::with_capacity(slice.len());
-
-                let mut i = 0;
-                for cur_char in slice.chars() {
-                    let size = cur_char.len_utf8();
-                    let bytes = slice[i..i + size].as_bytes();
-                    i += size;
-                    chars.extend(
-                        bytes
-                            .iter()
-                            .enumerate()
-                            .map(|(i, b)| (BYTES_CHAR[b], if i > 0 { 1 } else { 0 })),
-                    );
-                }
-
-                chars
-            })
-            .collect::<Vec<_>>();
-
-        // Update the NormalizedString
-        normalized.transform(
-            splits
-                .iter()
-                .flatten()
-                .map(|(c, changes)| (*c, *changes as isize)),
-            0,
-        );
-
-        // Collect splits and their offsets
-        let mut total_len = 0;
-        Ok(splits
-            .into_iter()
-            .map(|s| {
-                let mut len = 0;
-                let s = s
-                    .into_iter()
-                    .map(|(c, _)| {
-                        len += 1;
-                        c
-                    })
-                    .collect::<String>();
-                total_len += len;
-                (s, (total_len - len, total_len))
-            })
-            .collect())
+    fn pre_tokenize(&self, pretokenized: &mut PreTokenizedString) -> Result<()> {
+        let re_ref: &Regex = &RE;
+        pretokenized.split(|_, mut normalized| {
+            if self.add_prefix_space && !normalized.get().starts_with(' ') {
+                normalized.prepend(" ");
+            }
+            normalized.split(re_ref, SplitDelimiterBehavior::Isolated)
+        })?;
+        pretokenized.normalize(|normalized| {
+            let s = normalized.get();
+            let mut transformations: Vec<(char, isize)> = Vec::with_capacity(s.len());
+            let mut i = 0;
+            for cur_char in s.chars() {
+                let size = cur_char.len_utf8();
+                let bytes = s[i..i + size].as_bytes();
+                i += size;
+                transformations.extend(
+                    bytes
+                        .iter()
+                        .enumerate()
+                        .map(|(i, b)| (BYTES_CHAR[b], if i > 0 { 1 } else { 0 })),
+                );
+            }
+            normalized.transform(transformations.into_iter(), 0);
+            Ok(())
+        })
     }
 }
 
 /// As a `Decoder`, `ByteLevel` is in charge of converting any byte-level characters to their
 /// unicode counterpart, before merging everything back into a single String.
-#[typetag::serde]
 impl Decoder for ByteLevel {
     fn decode(&self, tokens: Vec<String>) -> Result<String> {
         let toks = tokens
@@ -172,7 +139,6 @@ impl Decoder for ByteLevel {
 }
 
 /// As a `PostProcessor`, `ByteLevel` is in charge of trimming the offsets if necessary.
-#[typetag::serde]
 impl PostProcessor for ByteLevel {
     fn added_tokens(&self, _is_pair: bool) -> usize {
         0
@@ -205,39 +171,31 @@ impl PostProcessor for ByteLevel {
 }
 
 pub fn process_offsets(encoding: &mut Encoding, add_prefix_space: bool) {
-    let modifs = encoding
-        .get_tokens()
-        .iter()
-        .map(|token| {
-            let leading_spaces = token
-                .chars()
-                .take_while(|c| *c == BYTES_CHAR[&b' '] || c.is_whitespace())
-                .count();
-            let trailing_spaces = token
-                .chars()
-                .rev()
-                .take_while(|c| *c == BYTES_CHAR[&b' '] || c.is_whitespace())
-                .count();
-            (leading_spaces, trailing_spaces)
-        })
-        .enumerate()
-        .filter(|(_, v)| v.0 > 0 || v.1 > 0)
-        .collect::<Vec<_>>();
+    encoding.process_tokens_with_offsets_mut(|(i, (token, mut offsets))| {
+        let mut leading_spaces = token
+            .chars()
+            .take_while(|c| *c == BYTES_CHAR[&b' '] || c.is_whitespace())
+            .count();
+        let trailing_spaces = token
+            .chars()
+            .rev()
+            .take_while(|c| *c == BYTES_CHAR[&b' '] || c.is_whitespace())
+            .count();
 
-    modifs.into_iter().for_each(|(i, (mut ld, tl))| {
-        let mut offsets = &mut encoding.get_offsets_mut()[i];
-        if ld > 0 {
-            if i == 0 && add_prefix_space && ld == 1 {
-                // If we are processing the first pair of offsets, with `add_prefix_space`,
-                // then we shouldn't remove anything we added. If there are more than one
-                // leading spaces though, it means we didn't add them, and they should be
-                // removed.
-                ld = 0;
+        if leading_spaces > 0 || trailing_spaces > 0 {
+            if leading_spaces > 0 {
+                if i == 0 && add_prefix_space && leading_spaces == 1 {
+                    // If we are processing the first pair of offsets, with `add_prefix_space`,
+                    // then we shouldn't remove anything we added. If there are more than one
+                    // leading spaces though, it means we didn't add them, and they should be
+                    // removed.
+                    leading_spaces = 0;
+                }
+                offsets.0 = std::cmp::min(offsets.0 + leading_spaces, offsets.1);
+            }
+            if trailing_spaces > 0 && offsets.1 >= trailing_spaces {
+                offsets.1 = std::cmp::max(offsets.1 - trailing_spaces, offsets.0);
             }
-            offsets.0 = std::cmp::min(offsets.0 + ld, offsets.1);
-        }
-        if tl > 0 && offsets.1 >= tl {
-            offsets.1 = std::cmp::max(offsets.1 - tl, offsets.0);
         }
     });
 }
@@ -246,26 +204,32 @@ pub fn process_offsets(encoding: &mut Encoding, add_prefix_space: bool) {
 mod tests {
     use super::ByteLevel;
     use crate::tokenizer::{
-        Decoder, Encoding, NormalizedString, PostProcessor, PreTokenizer, Range,
+        Decoder, Encoding, OffsetReferential, OffsetType, PostProcessor, PreTokenizedString,
+        PreTokenizer,
     };
 
     #[test]
     fn pre_tokenization() {
         let bytelevel = ByteLevel::default().add_prefix_space(false);
-        let mut input = NormalizedString::from("Hello my friend, how is your day going?");
+        let mut pretokenized: PreTokenizedString = "Hello my friend, how is your day going?".into();
+        bytelevel.pre_tokenize(&mut pretokenized).unwrap();
         assert_eq!(
-            bytelevel.pre_tokenize(&mut input).unwrap(),
+            pretokenized
+                .get_splits(OffsetReferential::Original, OffsetType::Byte)
+                .into_iter()
+                .map(|(s, o, _)| (s, o))
+                .collect::<Vec<_>>(),
             vec![
-                ("Hello".into(), (0, 5)),
-                ("Ġmy".into(), (5, 8)),
-                ("Ġfriend".into(), (8, 15)),
-                (",".into(), (15, 16)),
-                ("Ġhow".into(), (16, 20)),
-                ("Ġis".into(), (20, 23)),
-                ("Ġyour".into(), (23, 28)),
-                ("Ġday".into(), (28, 32)),
-                ("Ġgoing".into(), (32, 38)),
-                ("?".into(), (38, 39))
+                ("Hello", (0, 5)),
+                ("Ġmy", (5, 8)),
+                ("Ġfriend", (8, 15)),
+                (",", (15, 16)),
+                ("Ġhow", (16, 20)),
+                ("Ġis", (20, 23)),
+                ("Ġyour", (23, 28)),
+                ("Ġday", (28, 32)),
+                ("Ġgoing", (32, 38)),
+                ("?", (38, 39))
             ]
         );
     }
@@ -296,22 +260,25 @@ mod tests {
             " Hello my friend, how is your day going?",
             "Hello my friend, how is your day going?",
         ] {
-            let mut normalized = NormalizedString::from(s);
-            let pretok = bytelevel.pre_tokenize(&mut normalized).unwrap();
-            assert_eq!(normalized.get(), "ĠHelloĠmyĠfriend,ĠhowĠisĠyourĠdayĠgoing?");
+            let mut pretokenized = PreTokenizedString::from(*s);
+            bytelevel.pre_tokenize(&mut pretokenized).unwrap();
             assert_eq!(
-                pretok,
+                pretokenized
+                    .get_splits(OffsetReferential::Normalized, OffsetType::Byte)
+                    .into_iter()
+                    .map(|(s, o, _)| (s, o))
+                    .collect::<Vec<_>>(),
                 vec![
-                    ("ĠHello".into(), (0, 6)),
-                    ("Ġmy".into(), (6, 9)),
-                    ("Ġfriend".into(), (9, 16)),
-                    (",".into(), (16, 17)),
-                    ("Ġhow".into(), (17, 21)),
-                    ("Ġis".into(), (21, 24)),
-                    ("Ġyour".into(), (24, 29)),
-                    ("Ġday".into(), (29, 33)),
-                    ("Ġgoing".into(), (33, 39)),
-                    ("?".into(), (39, 40))
+                    ("ĠHello", (0, 7)),
+                    ("Ġmy", (7, 11)),
+                    ("Ġfriend", (11, 19)),
+                    (",", (19, 20)),
+                    ("Ġhow", (20, 25)),
+                    ("Ġis", (25, 29)),
+                    ("Ġyour", (29, 35)),
+                    ("Ġday", (35, 40)),
+                    ("Ġgoing", (40, 47)),
+                    ("?", (47, 48))
                 ]
             );
         }
@@ -327,11 +294,12 @@ mod tests {
 
         let bytelevel = ByteLevel::default().add_prefix_space(false);
         for sample in samples {
-            let mut input = NormalizedString::from(sample);
-            let pre_tokenized = bytelevel.pre_tokenize(&mut input).unwrap();
-            let separated_tokens = pre_tokenized
+            let mut pretokenized = PreTokenizedString::from(sample);
+            bytelevel.pre_tokenize(&mut pretokenized).unwrap();
+            let separated_tokens = pretokenized
+                .get_splits(OffsetReferential::Original, OffsetType::Byte)
                 .iter()
-                .flat_map(|(token, _)| token.split("").map(|t| t.into()))
+                .flat_map(|(s, _, _)| s.split("").map(|t| t.into()))
                 .collect::<Vec<_>>();
             assert_eq!(sample, bytelevel.decode(separated_tokens).unwrap());
         }
@@ -339,55 +307,78 @@ mod tests {
 
     #[test]
     fn handling_of_newlines() {
-        let mut input = NormalizedString::from("Hello there\nHello there");
+        let mut pretokenized = PreTokenizedString::from("Hello there\nHello there");
         let bytelevel = ByteLevel::default().add_prefix_space(false);
-        let p = bytelevel.pre_tokenize(&mut input).unwrap();
+        bytelevel.pre_tokenize(&mut pretokenized).unwrap();
 
         assert_eq!(
-            p,
+            pretokenized
+                .get_splits(OffsetReferential::Original, OffsetType::Byte)
+                .into_iter()
+                .map(|(s, o, _)| (s, o))
+                .collect::<Vec<_>>(),
             vec![
-                ("Hello".into(), (0, 5)),
-                ("Ġthere".into(), (5, 11)),
-                ("Ċ".into(), (11, 12)),
-                ("Hello".into(), (12, 17)),
-                ("Ġthere".into(), (17, 23))
+                ("Hello", (0, 5)),
+                ("Ġthere", (5, 11)),
+                ("Ċ", (11, 12)),
+                ("Hello", (12, 17)),
+                ("Ġthere", (17, 23))
             ]
         );
     }
 
     #[test]
     fn handling_of_multiple_whitespaces() {
-        let mut input = NormalizedString::from("Hello there       dear");
+        let mut pretokenized = PreTokenizedString::from("Hello there       dear");
         let bytelevel = ByteLevel::default().add_prefix_space(false);
-        let p = bytelevel.pre_tokenize(&mut input).unwrap();
+        bytelevel.pre_tokenize(&mut pretokenized).unwrap();
 
         assert_eq!(
-            p,
+            pretokenized
+                .get_splits(OffsetReferential::Original, OffsetType::Byte)
+                .into_iter()
+                .map(|(s, o, _)| (s, o))
+                .collect::<Vec<_>>(),
             vec![
-                ("Hello".into(), (0, 5)),
-                ("Ġthere".into(), (5, 11)),
-                ("ĠĠĠĠĠĠ".into(), (11, 17)),
-                ("Ġdear".into(), (17, 22))
+                ("Hello", (0, 5)),
+                ("Ġthere", (5, 11)),
+                ("ĠĠĠĠĠĠ", (11, 17)),
+                ("Ġdear", (17, 22))
             ]
         );
     }
 
     #[test]
     fn offsets_when_char_split_up() {
-        let mut input = NormalizedString::from("i⭢j");
+        let input = "i⭢j";
+        let mut pretokenized = PreTokenizedString::from(input);
         let bytelevel = ByteLevel::default().add_prefix_space(false);
-        let p = bytelevel.pre_tokenize(&mut input).unwrap();
+        bytelevel.pre_tokenize(&mut pretokenized).unwrap();
 
         assert_eq!(
-            p,
-            vec![
-                ("i".into(), (0, 1)),
-                ("âŃ¢".into(), (1, 4)),
-                ("j".into(), (4, 5)),
-            ]
+            pretokenized
+                .get_splits(OffsetReferential::Original, OffsetType::Byte)
+                .into_iter()
+                .map(|(s, o, _)| (s, o))
+                .collect::<Vec<_>>(),
+            vec![("i", (0, 1)), ("âŃ¢", (1, 4)), ("j", (4, 5))]
+        );
+        assert_eq!(
+            pretokenized
+                .get_splits(OffsetReferential::Normalized, OffsetType::Byte)
+                .into_iter()
+                .map(|(s, o, _)| (s, o))
+                .collect::<Vec<_>>(),
+            vec![("i", (0, 1)), ("âŃ¢", (1, 7)), ("j", (7, 8))]
+        );
+        assert_eq!(
+            pretokenized
+                .get_splits(OffsetReferential::Original, OffsetType::Byte)
+                .into_iter()
+                .map(|(_, o, _)| &input[o.0..o.1])
+                .collect::<Vec<_>>(),
+            vec!["i", "⭢", "j"]
         );
-        assert_eq!(input.get(), "iâŃ¢j");
-        assert_eq!(input.get_range_original(Range::Normalized(1..4)), Some("⭢"));
     }
 
     #[test]
