@@ -1,5 +1,8 @@
 use crate::tokenizer::{Encoding, PostProcessor, Result};
+use serde::ser::SerializeStruct;
+use serde::{Deserialize, Serialize, Serializer};
 
+#[derive(Deserialize, Clone, Debug)]
 pub struct BertProcessing {
     sep: (String, u32),
     cls: (String, u32),
@@ -135,3 +138,16 @@ impl PostProcessor for BertProcessing {
         Ok(new_encoding)
     }
 }
+
+impl Serialize for BertProcessing {
+    fn serialize<S>(&self, serializer: S) -> std::result::Result<S::Ok, S::Error>
+    where
+        S: Serializer,
+    {
+        let mut m = serializer.serialize_struct("BertProcessing", 3)?;
+        m.serialize_field("type", "BertProcessing")?;
+        m.serialize_field("sep", &self.sep)?;
+        m.serialize_field("cls", &self.cls)?;
+        m.end()
+    }
+}
